# Copyright (c) 2020 Watsen Networks.  All Rights Reserved.

_A='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2020-08-27",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-sztpd-1",\n        "revision": "2020-08-27",\n        "feature": [\n          "device-ownership-verification"\n        ],\n        "namespace": "https://watsen.net/sztpd-1",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
def nbi_0():return bss_common+bss_0_vZ
def nbi_1():return bss_common+bss_1_vZ
def nbi_x():return bss_common+bss_x_vZ
def nbi_x_tenant():return bss_common+bss_x_tenant_vZ
def sbi_rfc8572():return sbi_rfc8572_yl_str
def sbi_legacy():return'FIXME'
sbi_rfc8572_yl_str='\n{\n  "ietf-yang-library:modules-state": {\n    "module-set-id": "TBD",\n    "module": [\n      {\n        "name": "ietf-yang-types",\n        "revision": "2013-07-15",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-inet-types",\n        "revision": "2013-07-15",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-inet-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-netconf-acm",\n        "revision": "2018-02-14",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-netconf-acm",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-datastores",\n        "revision": "2018-02-14",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-datastores",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-yang-library",\n        "revision": "2019-01-04",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-library",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-sztp-bootstrap-server",\n        "revision": "2019-04-30",\n        "feature": [\n          "onboarding-server"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-sztp-bootstrap-server",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-yang-structure-ext",\n        "revision": "2019-03-07",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-structure-ext",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-crypto-types",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-crypto-types",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-sztp-csr",\n        "revision": "2020-06-02",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-sztp-csr",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}'
bss_common='\n{\n  "ietf-yang-library:modules-state": {\n    "module-set-id": "TBD",\n    "module": [\n      {\n        "name": "ietf-yang-types",\n        "revision": "2013-07-15",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-inet-types",\n        "revision": "2013-07-15",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-inet-types",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-datastores",\n        "revision": "2018-02-14",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-datastores",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-yang-library",\n        "revision": "2019-01-04",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-yang-library",\n        "conformance-type": "import"\n      },\n      {\n        "name": "iana-crypt-hash",\n        "revision": "2014-08-06",\n        "namespace": "urn:ietf:params:xml:ns:yang:iana-crypt-hash",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-x509-cert-to-name",\n        "revision": "2014-12-10",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-x509-cert-to-name",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-restconf",\n        "revision": "2017-01-26",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-restconf",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-netconf-acm",\n        "revision": "2018-02-14",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-netconf-acm",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-sztp-conveyed-info",\n        "revision": "2019-04-30",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-sztp-conveyed-info",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-crypto-types",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-crypto-types",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "ietf-truststore",\n        "revision": "2020-08-27",\n        "feature": [\n          "certificates"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-truststore",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-keystore",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-keystore",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tcp-common",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tcp-common",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tcp-client",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tcp-client",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tcp-server",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tcp-server",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tls-common",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tls-common",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tls-client",\n        "revision": "2020-08-27",\n        "feature": [\n          "x509-certificate-auth",\n          "client-auth-config-supported"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tls-client",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-tls-server",\n        "revision": "2020-08-27",\n        "feature": [\n          "x509-certificate-auth",\n          "client-auth-config-supported"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-tls-server",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-http-client",\n        "revision": "2020-08-27",\n        "feature": [\n          "basic-auth"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-http-client",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-http-server",\n        "revision": "2020-08-27",\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-http-server",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-restconf-client",\n        "revision": "2020-08-27",\n        "feature": [\n          "https-initiate"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-restconf-client",\n        "conformance-type": "import"\n      },\n      {\n        "name": "ietf-restconf-server",\n        "revision": "2020-08-27",\n        "feature": [\n          "http-listen",\n          "https-listen"\n        ],\n        "namespace": "urn:ietf:params:xml:ns:yang:ietf-restconf-server",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-x509-c2n",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/wnc2n",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-app-rpcs",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/app-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-app",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/wnapp",\n        "feature": [\n          "webhook-based-callouts",\n          "http-based-webhooks"\n        ],\n        "conformance-type": "import"\n      },\n'
bss_0_vZ='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2020-08-27",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
bss_1_vZ=_A
bss_x_vZ='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2020-08-27",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-sztpd-1",\n        "revision": "2020-08-27",\n        "feature": [\n          "device-ownership-verification"\n        ],\n        "namespace": "https://watsen.net/sztpd-1",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-sztpd-x",\n        "revision": "2020-08-27",\n        "feature": [\n          "device-ownership-verification"\n        ],\n        "namespace": "https://watsen.net/sztpd-x",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
bss_x_tenant_v1='\n      {\n        "name": "wn-sztpd-rpcs",\n        "revision": "2020-08-27",\n        "namespace": "https://watsen.net/sztpd-rpcs",\n        "conformance-type": "implement"\n      },\n      {\n        "name": "wn-sztpd-0",\n        "revision": "2020-08-27",\n        "feature": [\n          "onboarding-supported"\n        ],\n        "namespace": "https://watsen.net/sztpd-0",\n        "conformance-type": "import"\n      },\n      {\n        "name": "wn-sztpd-1",\n        "revision": "2020-08-27",\n        "feature": [\n          "system-level-administration-disabled"\n        ],\n        "namespace": "https://watsen.net/sztpd-1",\n        "conformance-type": "implement"\n      }\n    ]\n  }\n}\n'
bss_x_tenant_vZ=_A